/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JEntitySwitch.java,v 1.27.2.5 2003/12/30 08:16:19 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.util.HashMap;
import javax.ejb.EJBException;
import javax.ejb.EntityBean;
import javax.ejb.ObjectNotFoundException;
import javax.ejb.TimedObject;
import javax.ejb.Timer;
import javax.ejb.TimerService;
import javax.ejb.TransactionRolledbackLocalException;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JEntitySwitch is used internally to synchronize accesses to the entity context
 * and thus to the entity bean instance.
 * All parts common to EJBObject and EJBLocalObject should be here.
 * Different policies can be applied to manage context/instance pairs:
 * - only 1 pair (container manages the transaction isolation)
 * - 1 pair for each transaction (transaction isolation managed by DataBase)
 * - 2 pairs (1 for transactional accesses, 1 for non transaction accesses)
 * @author Philippe Durieux
 * @author Philippe Coq
 */
public class JEntitySwitch {

    /**
     * The Factory for this bean
     */
    protected JEntityFactory bf;

    /**
     * The Primary Key for this bean instance.
     */
    private Object pk = null;

    /**
     * The EJBLocalObject, or null if bean has no local interface.
     */
    private JEntityLocal local = null;

    /**
     * The EJBObject, or null if bean has no remote interface.
     */
    private JEntityRemote remote = null;

    /**
     * shared=true if the bean can be modify outside this container.
     */
    private boolean shared;

    /**
     * reentrant=true if a bean instance can be accessed concurrently.
     * this is only pertinent outside transactions.
     */
    private boolean reentrant;

    /**
     * Lock policy used for this entity bean.
     * Possible values are
     * 0 = LOCK_CONTAINER_READ_UNCOMMITTED (1 instance)
     * 1 = LOCK_CONTAINER_SERIALIZED (1 instance)
     * 2 = LOCK_CONTAINER_READ_COMMITTED (2 instances)
     * 3 = LOCK_DATABASE (n instances)
     */
    private int lockpolicy;

    /**
     * True if a transaction is mandatory for all modifying methods. In this case,
     * all methods outside tranction is considered as read-only.
     */
    private boolean txUpdates;

    /**
     * The Timer Service
     */
    private TimerService myTimerService = null;

    /**
     * nb of non transacted requests running
     */
    private int countIH = 0;

    /**
     * EntityContext for transacted requests when lockpolicy = 2,
     * or unique EntityContext when lockpolicy = 1.
     */
    private JEntityContext itContext = null;

    /**
     * EntityContext for non-transacted requests when lockpolicy = 2
     */
    private JEntityContext ihContext = null;

    /**
     * Map of EntityContext's for transactions when lockpolicy = 3
     * Key is the Transaction.
     */
    private HashMap itsContext;

    /**
     * nb of threads waiting (synchronization)
     */
    private int waiters = 0;

    /**
     * True if this instance may have been modified outside transactions.
     * Avoids to put it twice in dirty list.
     */
    private boolean inDirtyList = false;

    /**
     * True if a TX need this instance currently used outside tx
     * When the last release is done, we must store this instance.
     */
    private boolean mustStore = false;

    /**
     * True if instance has been modified by a Transaction.
     * This means that reading outside transaction should reload it before use.
     */
    private boolean mustReload = false;

    /**
     * True when context/instance has been discarded, to avoids that
     * passivate store its state on storage
     * Meaningful only outside tx
     */
    private boolean discarded = false;

    // transaction on which a Context has been registered.
    private Transaction runningtx = null;

    /**
     * last transaction blocked by synchronization.
     * this is to do a very basic deadlock detection.
     */
    private Transaction blockedtx = null;

    // max time to wait when in a transaction, before checking if
    // transaction has been set rollback only (deadlock detection)
    private static int maxtime = 3000; // 3 sec.

    // ident is used mainly for debugging
    private static int counter = 0;
    private String ident;

    /**
     * empty constructor. Object is initialized via init() because it is
     * implemented differently according to jorm mappers.
     */
    public JEntitySwitch() {
    }

    /**
     * constructor.
     * A new object is build when a new PK is known in the container,
     * either when a new bean is created, or when a find occurs.
     * For create(), PK is not known yet when this object is build.
     * @param bf The Entity Factory
     * @param pk The Primary Key
     */
    public void init(JEntityFactory bf, Object pk) {
        this.bf = bf;
        this.pk = pk;
        if (pk == null) {
            TraceEjb.logger.log(BasicLevel.ERROR, ident + "Init Entity Switch with a null PK!");
            throw new EJBException("Init Entity Switch with a null PK!");
        }

        // keep these in local, to be more efficient.
        shared = bf.isShared();
        reentrant = bf.isReentrant();
        lockpolicy = bf.lockPolicy();

        // Assume that all policies except "serialized" need transactional updates
        // Maybe possible to be less restrictive (LATER)
        txUpdates = (lockpolicy != 1);

        if (txUpdates) {
            reentrant = true;
        }
        if (lockpolicy > 2) {
            itsContext = new HashMap();
        }

        // generate new ident (debug)
        ident = "<" + counter + ":" + pk + ">";
        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident);
        counter++;

        countIH = 0;
        waiters = 0;

        // Create EJBObject if bean has a Remote Interface
        if (bf.getHome() != null) {
            try {
                remote = ((JEntityHome) bf.getHome()).createRemoteObject();
                remote.setEntitySwitch(this);
            } catch (RemoteException e) {
                throw new EJBException("cannot create Remote Object", e);
            }
        }

        // Create EJBLocalObject if bean has a Local Interface
        if (bf.getLocalHome() != null) {
            local = ((JEntityLocalHome) bf.getLocalHome()).createLocalObject();
            local.setEntitySwitch(this);
        }
    }

    /**
     * @param tx The Transaction
     * @return the JEntityContext used for this tx
     */
    private JEntityContext getContext4Tx(Transaction tx) {
        if (lockpolicy <= 1) {
            return itContext;
        }
        if (tx == null ) {
            return ihContext;
        } else {
            if (lockpolicy == 2) {
                return itContext;
            } else {
                return (JEntityContext) itsContext.get(tx);
            }
        }
    }

    /**
     * @param tx The Transaction
     * @param the JEntityContext used for this tx
     */
    private void setContext4Tx(Transaction tx, JEntityContext ctx) {
        if (lockpolicy <= 1) {
            itContext = ctx;
        } else {
            if (tx == null) {
                ihContext = ctx;
            } else {
                if (lockpolicy == 2) {
                    itContext = ctx;
                } else {
                    itsContext.put(tx, ctx);
                }
            }
        }
    }

    /**
     *
     */
    private void removeContext4Tx(Transaction tx) {
        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " get rid of this Context");
        // reset these flags
        discarded = false;
        // free this Context
        if (lockpolicy <= 1) {
            itContext = null;
        } else {
            if (tx == null) {
                ihContext = null;
            } else {
                if (lockpolicy == 2) {
                    itContext = null;
                } else {
                    itsContext.remove(tx);
                }
            }
        }
    }

    /**
     * @return the underlaying EJBLocalObject
     */
    public JEntityLocal getLocal() {
        return local;
    }

    /**
     * @return the underlaying EJBObject
     */
    public JEntityRemote getRemote() {
        return remote;
    }

    /**
     * Obtains the TimerService associated for this Entity Bean (one / pk)
     * @return a JTimerService instance.
     */
    public TimerService getEntityTimerService() {
        if (myTimerService == null) {
            // TODO : Check that instance implements TimedObject ?
            myTimerService = new JTimerService(this);
        }
        return myTimerService;
    }

    /**
     * Notify a timeout for this bean and this Pk
     * @param timer timer whose expiration caused this notification.
     */
    public void notifyTimeout(Timer timer) {
        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident);

        // Making this method generic leads to some restrictions listed here after :
        // - Transactional attribute is enforced to TX_REQUIRES_NEW.
        // - security is not checked.
        // LATER: Make it non generic and move it in JEntityLocal.vm and JEntityRemote.vm

        RequestCtx rctx = bf.preInvoke(MethodDesc.TX_REQUIRES_NEW, "");
        try {
            JEntityContext bctx = getICtx(rctx.currTx);
            EntityBean eb = bctx.getInstance();
            if (eb instanceof TimedObject) {
                TimedObject instance = (TimedObject) eb;
                instance.ejbTimeout(timer);
            } else {
                throw new EJBException("The bean does not implement the `TimedObject` interface");
            }
        } catch (EJBException e) {
            rctx.sysExc = e;
            throw e;
        } catch (RuntimeException e) {
            rctx.sysExc = e;
            throw new EJBException("RuntimeException thrown by an enterprise Bean", e);
        } catch (Error e) {
            rctx.sysExc = e;
            throw new EJBException("Error thrown by an enterprise Bean" + e);
        } catch (RemoteException e) {
            rctx.sysExc = e;
            throw new EJBException("Remote Exception raised:", e);
        } finally {
            try {
                bf.postInvoke(rctx);
            } finally {
                if (rctx.sysExc != null) {
                    discardICtx(rctx.currTx);
                } else {
                    releaseICtx(rctx.currTx);
                }
            }
        }
	}

    /**
     * @return the Primary Key Object for this instance.
     */
    public Object getPrimaryKey() {
        if (pk == null) {
            throw new EJBException("No Primary Key yet");
        }
        return pk;
    }

    /**
     * bind a JEntityContext for a create method.
     * No need to synchronize since object was unknown before. (TODO ?)
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind
     */
    public synchronized void bindICtx(Transaction tx, JEntityContext bctx) {
        mapICtx(tx, bctx, true, false);
    }

    /**
     * Try to bind a JEntityContext if none already bound.
     * Called by finder methods
     * @param tx - the Transaction object
     * @param bctx The Entity Context
     */
    public synchronized void tryBindICtx(Transaction tx, JEntityContext bctx) throws ObjectNotFoundException {

        // Don't bind if already a context.
        if (getContext4Tx(tx) != null) {
            TraceEjb.context.log(BasicLevel.DEBUG, ident + "context already mapped!");
            bf.releaseJContext(bctx);
            if (getContext4Tx(tx).isMarkedRemoved()) {
                TraceEjb.context.log(BasicLevel.DEBUG, ident + "Removed in the current tx");
                throw new ObjectNotFoundException("Removed in the current tx");
            }
            return;
        }

        // No synchronization, since this is used for readonly methods (finder);
        bctx.initEntityContext(this);
        bctx.activate(true);
        setContext4Tx(tx, bctx); // after activate (CRU => full parallellism)
        if (lockpolicy == 1 || lockpolicy == 2) {
            if (tx == null) {
                if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "IH find");
            } else {
                if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "IT find");
                registerCtx(tx, bctx);
            }
        }
    }

    /**
     * bind a JEntityContext for a remove method.
     * called in case of remove(pk) or remove(handle)
     * @param tx - the Transaction object
     * @param newctx - the JEntityContext to bind
     * @return the BeanContext
     */
    public JEntityContext getICtx(Transaction tx, JEntityContext newctx) {
        return mapICtx(tx, newctx, false, false);
    }

    /**
     * Get a context/instance associated with this transaction
     * Called at each request on the bean (including remove)
     * @param tx - the Transaction object
     * @return the BeanContext
     */
    public JEntityContext getICtx(Transaction tx) {
        return mapICtx(tx, null, false, true);
    }

    /**
     * Map a context and its instance.
     * This version manages only 1 context (iContext).
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind if not null
     * @param forced - force to take this context. (case of create)
     * @param holdit - increment count to hold it, a release will be called later.
     * @return JEntityContext actually mapped
     */
    public synchronized JEntityContext mapICtx(Transaction tx, JEntityContext bctx, boolean forced, boolean holdit) {

        // Synchronization.
        if (tx == null) {
            // ------------- IH begin -------------
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH");

            // Must wait in case of TX or if non reentrant and already busy
            while ((runningtx != null && lockpolicy == 1) || (! reentrant && countIH > 0)) {
                if (runningtx != null) {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH: WAIT end IT");
                } else {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH: WAIT end IH");
                }
                waiters++;
                try {
                    wait(maxtime);
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH: NOTIFIED");
                } catch (InterruptedException e) {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH: INTERRUPTED");
                } catch (Exception e) {
                    throw new EJBException("JEntitySwitch synchronization pb", e);
                } finally {
                    waiters--;
                }
            }
        } else {
            // ------------- IT begin -------------
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT");

            // Must wait in case of TX or if instance has been modified outside transaction.
            // Don't wait transactions if 1 instance per transaction.
            int waitcount = 0;
            while (inDirtyList || ((lockpolicy == 1 || lockpolicy == 2) && runningtx != null && !tx.equals(runningtx))) {
                // instance is dirty : must write it before working on it again.
                // this is mandatory to be able to retrieve a good state in case of rollback.
                if (inDirtyList) {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT: WAIT end IH");
                    // Cannot passivate here, because must be done outside tx context
                    bf.synchronizeEntities();
                } else {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT: WAIT end IT");
                    // deadlock detection
                    blockedtx = tx;
                    if (waitcount++ > 1 && bf.isBlocked(runningtx) && bf.isBlocking(tx)) {
                        try {
                            tx.setRollbackOnly();
                        } catch (SystemException e) {
                            TraceEjb.logger.log(BasicLevel.ERROR, ident + "getICtx IT: unexpected exception setting rollbackonly");
                        }
                        TraceEjb.logger.log(BasicLevel.WARN, ident + "getICtx IT: transaction rolled back");
                        throw new TransactionRolledbackLocalException("possible deadlock");
                    }
                }
                waiters++;
                try {
                    wait(maxtime);
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT: NOTIFIED");
                } catch (InterruptedException e) {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT: INTERRUPTED");
                } catch (Exception e) {
                    throw new EJBException("JEntitySwitch synchronization pb", e);
                } finally {
                    waiters--;
                    blockedtx = null;
                }
                // If transaction has been rolledback or set rollback only, give up.
                int status = Status.STATUS_ROLLEDBACK;
                try {
                    status = tx.getStatus();
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "getICtx IT: unexpected exception getting transaction status");
                }
                switch (status) {
                case Status.STATUS_MARKED_ROLLBACK:
                case Status.STATUS_ROLLEDBACK:
                case Status.STATUS_ROLLING_BACK:
                    TraceEjb.logger.log(BasicLevel.WARN, ident + "getICtx IT: transaction rolled back");
                    throw new TransactionRolledbackLocalException("rollback occured while waiting");
                }
            }
        }

        // Choose the context to use.
        boolean newtrans = false;
        JEntityContext jec = getContext4Tx(tx);
        if (forced) {
            // If the new context is enforced, we must first release the older
            if (jec != null) {
                if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "new context is enforced!");
                bf.releaseJContext(jec);
                removeContext4Tx(tx);
            }
            jec = bctx;
            setContext4Tx(tx, jec);
            jec.initEntityContext(this);
            newtrans = true;
        } else {
            if (jec != null) {
                // Reuse the Context for this transaction.
                // If a context was supplied, release it first.
                if (bctx != null) {
                    if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "a context was supplied!");
                    bf.releaseJContext(bctx);
                }
                jec.reuseEntityContext();
                // In case the same instance is used for all tx, must check if new one.
                if (runningtx == null && (lockpolicy == 1 || lockpolicy == 2)) {
                    newtrans = true;
                }
            } else {
                if (bctx != null) {
                    jec = bctx;
                } else {
                    // no Context available : get one from the pool.
                    jec = (JEntityContext) bf.getJContext();
                }
                jec.initEntityContext(this);
                jec.activate(true);
                setContext4Tx(tx, jec); // after activate
                newtrans = true;
            }
        }

        if (tx != null) {
            if (newtrans && (lockpolicy == 1 || lockpolicy == 2)) {
                if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT: new tx, registerSynchronization");
                registerCtx(tx, jec);
            }
        } else {
            if (holdit) {
                countIH++;
                if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH count=" + countIH);
                if ((shared || mustReload) && countIH == 1) {
                    // reload state that could have been modified by transactions.
                    jec.activate(false);
                    mustReload = false;
                }
            }
            if (!inDirtyList && !txUpdates) {
                setDirty();
            }
        }

        return jec;
    }

    public Transaction getBlockedTx() {
        return blockedtx;
    }

    public Transaction getBlockingTx() {
        return runningtx;
    }

    /**
     * Release a context/instance at end of request.
     * @param tx - transaction associated to this context
     */
    public synchronized void releaseICtx(Transaction tx) {

        if (tx == null) {
            // ------------- IH end -------------
            JEntityContext jec = getContext4Tx(tx);
            countIH--;
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "releaseICtx IH count=" + countIH);
            if (countIH == 0) {
                if (discarded) {
                    if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "discarded!");
                    bf.releaseJContext(jec);
                    removeContext4Tx(tx);
                } else {
                    if (jec.isMarkedRemoved()) {
                        if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "Marked Removed!");
                        bf.releaseJContext(jec);
                        removeContext4Tx(tx);
                        noLongerUsed();
                    } else if (mustStore) {
                        try {
                            jec.storeIfModified();
                        } catch (EJBException e) {
                            TraceEjb.logger.log(BasicLevel.WARN, ident + " ejbexception:" + e);
                        }
                        mustStore = false;
                    }
                }
                if (waiters > 0) {
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
                    notifyAll();
                }
            }
        } else {
            // ------------- IT end -------------
            // nothing to do now. Wait for tx completed.
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "releaseICtx IT");

        }
    }

    /**
     * Discard a context/instance at end of request. A problem occured
     * on this insance and it must be discarded.
     * @param tx - transaction associated to this context
     */
    public synchronized void discardICtx(Transaction tx) {
        if (tx == null) {
            JEntityContext jec = getContext4Tx(tx);
            countIH--;
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "discardICtx IH count=" + countIH);
            if (countIH == 0) {
                if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "discardICtx!");
                bf.releaseJContext(jec);
                removeContext4Tx(tx);
            } else {
                discarded = true;
            }
        } else {
            // nothing to do now. Wait for tx completed.
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "discardICtx IT");
        }
    }

    /**
     * This transaction is now over. We can dispose of the instance
     * for another transaction or discard it.
     * @param tx the transaction object
     * @param committed true if transaction was committed.
     */
    public synchronized void txCompleted(Transaction tx, boolean committed) {
        JEntityContext jec = getContext4Tx(tx);
        runningtx = null;

        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "txCompleted " + committed);

        // The create was eventually rolled back. We must remove PK entry.
        if (jec.isNewInstance() && !committed) {
            if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "create rolled back!");
            bf.releaseJContext(jec);
            removeContext4Tx(tx);
            noLongerUsed();
            return;
        }
            
        if (jec.isMarkedRemoved()) {
            if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "remove!");
            bf.releaseJContext(jec);
            removeContext4Tx(tx);
            if (committed) {
                noLongerUsed(); // remove is effective !
            }
        } else {
            if (shared || !committed) {
                if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, jec + " passivated!");
                jec.passivate();
                bf.releaseJContext(jec);
                removeContext4Tx(tx);
            } else {
                // do this before reallocating the Context for another tx (inside this lock)
                jec.detachTx();
            }
            // If special instance for no transactional accesses, it must be reloaded
            if (lockpolicy >= 2) {
                mustReload = true;
            }
        }
        if (waiters > 0) {
            TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
            notifyAll();
        }
    }

    /**
     * Instance is ready to use for new transaction.
     */
    public synchronized void endIH() {
        inDirtyList = false;
        if (getContext4Tx(null) == null) {
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " discarded!");
            return;
        }
        // If the instance is busy, don't touch it.
        // If no context mapped, nothing to do
        if (countIH == 0) {
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " ready again");
        } else {
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " busy!");
        }
        if (waiters > 0) {
            TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
            notifyAll();
        }
    }

    /**
     * This transaction has just modified this instance. (CMP2 only)
     * This is used for lockpolicy 0 or 3 only.
     * @param tx transaction
     */
    public synchronized void notifyWriting(Transaction tx, JEntityContext bctx) {
        TraceEjb.swapper.log(BasicLevel.DEBUG, ident);
        if (runningtx == null) {
            registerCtx(tx, bctx);
        } else {
            // debug
            if (runningtx == tx) {
                TraceEjb.swapper.log(BasicLevel.DEBUG, ident + " called twice ?");
                return;
            }
            TraceEjb.logger.log(BasicLevel.WARN, "Conflict: 2 transactions attempt to write the same instance");
            try {
                tx.setRollbackOnly();
            } catch (SystemException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "cannot setRollbackOnly");
            }
            if (lockpolicy == 0) {
                // Since we share the same instance, we must rollback also
                // the current transaction (else data would be corrupted).
                try {
                    runningtx.setRollbackOnly();
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, "cannot setRollbackOnly");
                }
                // We should also avoid modifying instance, in case runningtx is
                // already committing (setRollbackOnly will fail)
                throw new EJBException("Conflict writing entity bean");
            }
        }
    }

    /**
     * try to passivate IH (called from swapper)
     * @param passivation false if only store bean state.
     * @return true if OK, false if busy.
     */
    public synchronized boolean passivateIH(boolean passivation) {

        JEntityContext jec = getContext4Tx(null);

        // If no context mapped or readonly, nothing to do
        if (jec == null || txUpdates) {
            return true;
        }

        // If the instance is busy, don't touch it.
        if (countIH > 0) {
            // will be stored later, when released by last thread (countIH = 0)
            // return true to remove from the dirty list, since it will be written anyway.
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " used off TX");
            mustStore = true;
            return true;
        }
        if (runningtx != null && lockpolicy == 1) {
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " used in TX");
            return false;
        }

        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident);

        // Store the instance state
        // -> only if instance not discarded
        if (!discarded) {
            try {
                jec.storeIfModified();
            } catch (Exception e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ident, "error while storing bean state:", e);
            }
            mustStore = false;
            // don't set dirty false here, will be done in endIH.
            // this prevent duplicates in dirty list.
        }

        // passivate this instance if required.
        if (passivation) {
            if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, "passivated: " + jec);
            jec.passivate();
            bf.releaseJContext(jec);
            removeContext4Tx(null);
            if (waiters > 0) {
                TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
                notifyAll();
            }
        }
        return true;
    }

    /**
     * Remove this EJB and
     * UnExport the remote object.
     */
    private void noLongerUsed() {
        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident);
        if (remote != null) {
            try {
                remote.unexportObject();
            } catch (NoSuchObjectException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ident + " unexport entity failed: " + e);
            }
        }
        bf.removeEJB(getPrimaryKey());
    }

    /**
     * instance is dirty : possibly modified outside a transaction.
     */
    private void setDirty() {
        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG, ident);
        inDirtyList = true;
        bf.registerEJB(this);
    }

    /**
     * register a Context on the transaction, as a Synchronization.
     * this will be used later to store instance state when needed :
     * before a finder, or at beforeCompletion, and to release instance
     * at commit.
     * @param tx Transaction object
     * @param bctx The Context to be registered
     */
    private void registerCtx(Transaction tx, JEntityContext bctx) {
        // register this context to the transaction, as a Synchronization.
        try {
            TraceEjb.context.log(BasicLevel.DEBUG, bctx + "registered!");
            bf.registerContext(tx, bctx);
        } catch (IllegalStateException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ident + "Transaction is in an illegal state");
            throw e;
        }
        runningtx = tx;
        bctx.setRunningTx(tx);
    }

    /**
     * @return lock policy for this bean
     */
    public int getPolicy() {
        return lockpolicy;
    }


    /**
     * @return ejb instance's state
     * state values are 0=in-tx, 1=out-tx, 2=idle, 3=passive, 4=removed.
     * we don't synchronize this method to avoid jadmin blocks
     */
    public int getState() {
        JEntityContext jec = getContext4Tx(null); // TODO: change this null !
        if (jec == null) {
            return 3;
        }
        if (runningtx != null) {
            return 0;
        }
        if (jec.isMarkedRemoved()) {
            return 4;
        }
        if (inDirtyList) {
            return 1;
        } else {
            return 2;
        }
    }

}
